"""Tests for the results module."""
from alglab.results import Results


def test_plots():
    # Run a simple experiment
    results = Results("results/results.csv")
    assert results.num_runs == 2

    results.line_plot('noise', 'running_time_s')




"""
Create a generic class representing an algorithm which can be applied to a dataset.
"""
from typing import List, Type, Callable, Dict
import alglab.dataset


class Algorithm(object):

    def __init__(self,
                 name: str,
                 implementation: Callable,
                 return_type: Type = None,
                 parameter_names: List[str] = None,
                 dataset_class: Type[alglab.dataset.Dataset] = alglab.dataset.NoDataset):
        """Create an algorithm definition. The implementation should be a python method which takes
        a dataset as a positional argument (if dataset_class is not NoDataset) and
        the parameters as keyword arguments. The implementation should return an object of type
        return_type.
        """
        self.implementation = implementation
        self.parameter_names = parameter_names if parameter_names is not None else []
        self.return_type = return_type
        self.dataset_class = dataset_class
        self.name = name

    def run(self, dataset: alglab.dataset.Dataset, params: Dict):
        if not isinstance(dataset, self.dataset_class):
            raise TypeError("Provided dataset type must match dataset_class expected by the implementation.")

        for param in params.keys():
            if param not in self.parameter_names:
                raise ValueError("Unexpected parameter name.")

        if self.dataset_class is not alglab.dataset.NoDataset:
            result = self.implementation(dataset, **params)
        else:
            result = self.implementation(**params)

        if not isinstance(result, self.return_type):
            raise TypeError("Provided result type must match promised return_type.")

        return result

    def __repr__(self):
        return self.name



"""Tests for the experiment module."""
from sklearn.cluster import KMeans, SpectralClustering
import numpy as np
import alglab.dataset
import alglab.experiment
import alglab.evaluation
import alglab.algorithm


# We will use this KMeans implementation throughout the tests.
def kmeans_impl(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def sc_impl(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_sc = SpectralClustering(n_clusters=k)
    sklearn_sc.fit(data.data)
    return sklearn_sc.labels_


def test_experimental_suite():
    # Test the experimental suite class as it's intended to be used.
    alg1 = alglab.algorithm.Algorithm("kmeans",
                                     kmeans_impl,
                                     np.ndarray,
                                     ["k"],
                                     alglab.dataset.PointCloudDataset)
    alg2 = alglab.algorithm.Algorithm("sc",
                                     sc_impl,
                                     np.ndarray,
                                     ["k"],
                                     alglab.dataset.PointCloudDataset)

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        alg_fixed_params={'kmeans': {'k': 2}, 'sc': {'k': 2}},
        dataset_fixed_params={'n': 1000},
        dataset_varying_params={'noise': np.linspace(0, 1, 5)},
        evaluators=[alglab.evaluation.adjusted_rand_index]
        )
    experiments.run_all()


def test_multiple_runs():
    # Test the experimental suite class as it's intended to be used.
    alg1 = alglab.algorithm.Algorithm("kmeans",
                                     kmeans_impl,
                                     np.ndarray,
                                     ["k"],
                                     alglab.dataset.PointCloudDataset)
    alg2 = alglab.algorithm.Algorithm("sc",
                                     sc_impl,
                                     np.ndarray,
                                     ["k"],
                                     alglab.dataset.PointCloudDataset)

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        alg_fixed_params={'kmeans': {'k': 2}, 'sc': {'k': 2}},
        dataset_fixed_params={'n': 1000},
        dataset_varying_params={'noise': np.linspace(0, 1, 5)},
        evaluators=[alglab.evaluation.adjusted_rand_index],
        num_runs=2
    )
    assert experiments.num_trials == 20
    experiments.run_all()


def test_dynamic_params():
    alg1 = alglab.algorithm.Algorithm("kmeans",
                                     kmeans_impl,
                                     np.ndarray,
                                     ["k"],
                                     alglab.dataset.PointCloudDataset)
    alg2 = alglab.algorithm.Algorithm("sc",
                                     sc_impl,
                                     np.ndarray,
                                     ["k"],
                                     alglab.dataset.PointCloudDataset)

    experiments = alglab.experiment.ExperimentalSuite(
        [alg1, alg2],
        alglab.dataset.TwoMoonsDataset,
        "results/twomoonsresults.csv",
        alg_fixed_params={'kmeans': {'k': 2}},
        alg_varying_params={'sc': {'k': [(lambda p: int(p['n'] / 100)), 2]}},
        dataset_fixed_params={'noise': 0.1},
        dataset_varying_params={'n': np.linspace(100, 1000, 5).astype(int)},
        evaluators=[alglab.evaluation.adjusted_rand_index]
    )
    experiments.run_all()




"""Implementation of the Dataset object for use with algpy."""
from sklearn.datasets import make_moons, fetch_openml
from sklearn.neighbors import kneighbors_graph
import numpy as np
from abc import ABC, abstractmethod
import stag.graph
import stag.random
import matplotlib.pyplot as plt
from typing import Type, Dict
import pandas as pd


class Dataset(ABC):

    @abstractmethod
    def __init__(self, **kwargs):
        """Construct the dataset."""
        pass


class NoDataset(Dataset):
    """Use this when no dataset is needed to compare algorithms."""

    def __init__(self):
        pass

    def __str__(self):
        return "NoDataset"


class ClusterableDataset(Dataset):
    """
    A dataset which may have ground truth clusters.
    """

    def __init__(self, labels):
        self.gt_labels = labels


class GraphDataset(ClusterableDataset):
    """
    A dataset whose central data is a graph.
    """

    def __init__(self, graph: stag.graph.Graph = None, labels=None):
        """Initialise the dataset with a stag Graph. Optionally, provide ground truth
        labels for classification."""
        self.graph = graph
        self.n = 0 if graph is None else graph.number_of_vertices()
        ClusterableDataset.__init__(self, labels)


class SBMDataset(GraphDataset):
    """
    Create a graph dataset from a stochastic block model.
    """

    def __init__(self, n: int = 1000, k: int = 10, p: float = 0.5, q: float = 0.1):
        self.n = int(n)
        self.k = int(k)
        self.p = p
        self.q = q
        g = stag.random.sbm(self.n, self.k, p, q)
        labels = stag.random.sbm_gt_labels(self.n, self.k)
        GraphDataset.__init__(self, graph=g, labels=labels)


    def __repr__(self):
        return f"SBMDataset({self.n}, {self.k}, {self.p}, {self.q})"


class PointCloudDataset(ClusterableDataset):
    """
    The simplest form of dataset: the data consists of a point cloud in Euclidean space.
    This is represented internally by a numpy array.
    """

    def __init__(self, data: np.array = None, labels=None):
        """Initialise the dataset with a numpy array. Optionally, provide labels for classification."""
        self.data = np.array(data)
        self.n, self.d = data.shape
        ClusterableDataset.__init__(self, labels)

    def plot_clusters(self, labels):
        """
        If the data is two-dimensional, plot the data, colored according to the labels.
        """
        if self.d != 2:
            raise ValueError("Dataset is not two-dimensional.")

        if len(labels) != self.n:
            raise ValueError("Labels length must match number of data points.")

        labels = np.array(labels)

        # Plot the data points, colored by their cluster labels
        plt.figure(figsize=(10, 7))
        unique_labels = np.unique(labels)

        for label in unique_labels:
            cluster_data = self.data[labels == label]
            plt.scatter(cluster_data[:, 0], cluster_data[:, 1], label=f'Cluster {label}')

        plt.grid(True)
        plt.show()


class TwoMoonsDataset(PointCloudDataset):
    """The toy two moons dataset from sklearn."""

    def __init__(self, n=1000, noise=0.07):
        """Initialise the two moons dataset. Optionally, provide the number of points, n, and the noise parameter."""
        x, y = make_moons(n_samples=n, noise=noise)
        PointCloudDataset.__init__(self, data=x, labels=y)

    def __str__(self):
        return f"TwoMoonsDataset({self.n})"


class OpenMLDataset(PointCloudDataset):
    """Load pointcloud data from OpenML."""

    def __init__(self, **kwargs):
        """Initialise the dataset by downloading from openML. Accepts the same arguments as the
        sklearn fetch_openml method."""
        data_info = fetch_openml(**kwargs)
        if isinstance(data_info.data, pd.DataFrame):
            data_info.data = data_info.data.to_numpy()

        target = data_info.target
        if isinstance(target, pd.Series) or isinstance(target, pd.DataFrame):
            if isinstance(target.dtype, pd.CategoricalDtype):
                target = target.cat.codes.to_numpy()
            else:
                target = data_info.target.to_numpy()

        PointCloudDataset.__init__(self, data=data_info.data, labels=target)


class KnnGraphDataset(GraphDataset, PointCloudDataset):
    """A k-nearest neighbour graph dataset is both a point cloud and a graph dataset."""

    def __init__(self,
                 k: int = 10,
                 pointcloud_class: Type[PointCloudDataset] = PointCloudDataset,
                 **pointcloud_parameters):
        # Initialise this as a pointcloud dataset
        pointcloud_class.__init__(self, **pointcloud_parameters)

        # Create the k nearest neighbours graph and initialise as a graph dataset
        adj_non_symmetric = kneighbors_graph(self.data, k)
        g = stag.graph.Graph(adj_non_symmetric + adj_non_symmetric.transpose())
        GraphDataset.__init__(self, graph=g, labels=self.gt_labels)



"""Tests for the algorithm class."""
import pytest

import alglab.dataset
import alglab.algorithm
import alglab.evaluation
from sklearn.cluster import KMeans
import numpy as np
from typing import List


# We will use this KMeans implementation throughout the tests.
def kmeans_impl(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def no_dataset_impl(n=10):
    return n ** 3


def test_algorithm():
    # Test the algorithm class as it is intended to be used.
    kmeans_algorithm = alglab.algorithm.Algorithm('kmeans',
                                                 kmeans_impl,
                                                 return_type=np.ndarray,
                                                 dataset_class=alglab.dataset.PointCloudDataset,
                                                 parameter_names=['k'])

    test_data = alglab.dataset.TwoMoonsDataset()
    _ = kmeans_algorithm.run(test_data, {'k': 8})


def test_bad_return_type():
    # Specify the wrong return type
    kmeans_algorithm = alglab.algorithm.Algorithm('kmeans',
                                                 kmeans_impl,
                                                 return_type=List,
                                                 dataset_class=alglab.dataset.PointCloudDataset,
                                                 parameter_names=['k'])

    test_data = alglab.dataset.TwoMoonsDataset()

    with pytest.raises(TypeError, match='return_type'):
        _ = kmeans_algorithm.run(test_data, {'k': 8})


def test_bad_parameter_names():
    # Specify the wrong parameter names
    kmeans_algorithm = alglab.algorithm.Algorithm('kmeans',
                                                 kmeans_impl,
                                                 return_type=np.ndarray,
                                                 dataset_class=alglab.dataset.PointCloudDataset,
                                                 parameter_names=['m'])

    test_data = alglab.dataset.TwoMoonsDataset()

    with pytest.raises(ValueError, match='parameter'):
        _ = kmeans_algorithm.run(test_data, {'k': 8})


def test_unspecified_parameter():
    # Test the algorithm class as it is intended to be used.
    kmeans_algorithm = alglab.algorithm.Algorithm('kmeans',
                                                 kmeans_impl,
                                                 return_type=np.ndarray,
                                                 dataset_class=alglab.dataset.PointCloudDataset,
                                                 parameter_names=['k'])

    # Not fully specifying the parameter is permitted when running.
    test_data = alglab.dataset.TwoMoonsDataset()
    _ = kmeans_algorithm.run(test_data, {})


def test_no_dataset():
    cube_algorithm = alglab.algorithm.Algorithm('power',
                                               no_dataset_impl,
                                               return_type=int,
                                               parameter_names=['n'])

    _ = cube_algorithm.run(alglab.dataset.NoDataset(), {'n': 20})


def test_wrong_dataset():
    cube_algorithm = alglab.algorithm.Algorithm('power',
                                               no_dataset_impl,
                                               return_type=int,
                                               parameter_names=['n'])

    with pytest.raises(TypeError, match='dataset'):
        _ = cube_algorithm.run(alglab.dataset.TwoMoonsDataset(), {'n': 20})


def test_openml_dataset():
    dataset = alglab.dataset.OpenMLDataset(name="iris")
    kmeans_algorithm = alglab.algorithm.Algorithm('kmeans',
                                                  kmeans_impl,
                                                  return_type=np.ndarray,
                                                  dataset_class=alglab.dataset.PointCloudDataset,
                                                  parameter_names=['k'])

    labels = kmeans_algorithm.run(dataset, {'k': 3})

    # Check that evaluation works
    _ = alglab.evaluation.adjusted_rand_index.apply(dataset, labels)



"""
Methods for evaluating the performance of an algorithm.
"""
from typing import Callable, Type
import numpy as np
import alglab.algorithm
import alglab.dataset
import stag.cluster
import stag.graph
import scipy.sparse.linalg


class Evaluator(object):

    def __init__(self,
                 name: str,
                 implementation: Callable,
                 alg_result_type: Type = None,
                 dataset_class: Type[alglab.dataset.Dataset] = alglab.dataset.NoDataset):
        """Define a method of evaluating an algorithm. Specify the evaluator implementation as
        well as the expected result type of the algorithm to be evaluated and the type of the dataset
        the algorithm should be applied to."""
        self.implementation = implementation
        self.alg_result_type = alg_result_type
        self.dataset_class = dataset_class
        self.name = name

    def apply(self, dataset: alglab.dataset.Dataset, alg_result):
        if not isinstance(dataset, self.dataset_class):
            raise TypeError(f"Expected dataset type to be {self.dataset_class}, got {type(dataset)}.")

        if not isinstance(alg_result, self.alg_result_type):
            raise TypeError(f"Expected alg_result type to be {self.alg_result_type} but got {type(alg_result)}.")

        if self.dataset_class is not alglab.dataset.NoDataset:
            result = self.implementation(dataset, alg_result)
        else:
            result = self.implementation(alg_result)

        return result

    def __str__(self):
        return self.name

# -----------------------------------------------------------------------------
# Clustering Evaluation
# -----------------------------------------------------------------------------


def __ari_impl(data: alglab.dataset.ClusterableDataset, labels):
    if data.gt_labels is not None:
        return stag.cluster.adjusted_rand_index(data.gt_labels, labels)
    else:
        raise ValueError('No ground truth labels provided.')


adjusted_rand_index = Evaluator('adjusted_rand_index',
                                __ari_impl,
                                alg_result_type=np.ndarray,
                                dataset_class=alglab.dataset.ClusterableDataset)


# -----------------------------------------------------------------------------
# Graph Evaluation
# -----------------------------------------------------------------------------

def __num_vertices_impl(_: alglab.dataset.Dataset, graph: stag.graph.Graph):
    return graph.number_of_vertices()


num_vertices = Evaluator('number_of_vertices',
                         __num_vertices_impl,
                         alg_result_type=stag.graph.Graph,
                         dataset_class=alglab.dataset.Dataset)


def __avg_degree_impl(_: alglab.dataset.Dataset, graph: stag.graph.Graph):
    return graph.average_degree()


avg_degree = Evaluator('average_degree',
                       __avg_degree_impl,
                       alg_result_type=stag.graph.Graph,
                       dataset_class=alglab.dataset.Dataset)


def __normalised_laplacian_second_eigenvalue_impl(_: alglab.dataset.Dataset, graph: stag.graph.Graph):
    lap = graph.normalised_laplacian().to_scipy()
    eigs, _ = scipy.sparse.linalg.eigsh(lap, which='SM', k=2)
    return eigs[1]


normalised_laplacian_second_eigenvalue = Evaluator('lap_second_eigenvalue',
                                                   __normalised_laplacian_second_eigenvalue_impl,
                                                   alg_result_type=stag.graph.Graph,
                                                   dataset_class=alglab.dataset.Dataset)



"""Classes and methods related to running experiments on algorithms and datasets."""
import time

from typing import Dict, Type, List, Callable, Iterable
import pandas as pd
import itertools
from collections import OrderedDict

import alglab.algorithm
import alglab.dataset
import alglab.results
import alglab.evaluation


def product_dict(**kwargs):
    keys = kwargs.keys()
    for instance in itertools.product(*kwargs.values()):
        yield dict(zip(keys, instance))


def resolve_parameters(fixed_parameters, varying_parameters):
    """
    Given a dictionary of fixed parameters and a dictionary of varying parameters, resolve the varying parameters
    which are defined by function.
    """
    resolved_parameters = {}
    dynamic_parameters = []
    for param_name, value in varying_parameters.items():
        if not callable(value):
            # This is a 'static' parameter
            resolved_parameters[param_name] = value
        else:
            # This is a dynamically defined parameter
            dynamic_parameters.append(param_name)

    # Resolve the dynamic parameters
    for param_name in dynamic_parameters:
        resolved_parameters[param_name] = varying_parameters[param_name](fixed_parameters | resolved_parameters)

    return resolved_parameters


class Experiment(object):

    def __init__(self, alg: alglab.algorithm.Algorithm, dataset: alglab.dataset.Dataset, params,
                 evaluators: List[alglab.evaluation.Evaluator] = None):
        """An experiment is a single instance of running an algorithm on a dataset with a set of parameters.
        The running time of the algorithm is measured by default. In addition to this, the evaluation_functions
        variable should contain a dictionary of methods which will be applied to the result of the algorithm.
        """
        self.alg = alg
        self.dataset = dataset
        self.params = params
        self.evaluators = evaluators

        # The result will be a dictionary of information we want to track after each run.
        self.result = {}

    def run(self):
        """Run the experiment."""
        # We always measure the running time of the experiment.
        start_time = time.time()
        alg_output = self.alg.run(self.dataset, self.params)
        end_time = time.time()
        self.result['running_time_s'] = end_time - start_time

        # Apply the evaluation functions
        if self.evaluators:
            for evaluator in self.evaluators:
                self.result[evaluator.name] = evaluator.apply(self.dataset, alg_output)


class ExperimentalSuite(object):

    def __init__(self,
                 algorithms: List[alglab.algorithm.Algorithm],
                 dataset: Type[alglab.dataset.Dataset],
                 results_filename: str,
                 num_runs: int = 1,
                 alg_fixed_params: Dict[str, Dict] = None,
                 alg_varying_params:  Dict[str, Dict] = None,
                 dataset_fixed_params: Dict = None,
                 dataset_varying_params: Dict = None,
                 evaluators: List[alglab.evaluation.Evaluator] = None):
        """Run a suite of experiments while varying some parameters.

        Varying parameter dictionaries should have parameter names as keys and the values should be an iterable containing:
            - values to be used directly; or
            - functions, taking fixed and statically defined variable parameters and returning a parameter value
        """
        self.num_runs = num_runs

        if num_runs < 1:
            raise ValueError('num_runs must be greater than or equal to 1')

        self.algorithms = algorithms
        self.algorithm_names = [alg.name for alg in self.algorithms]

        # Automatically populate the parameter dictionaries
        if alg_fixed_params is None:
            alg_fixed_params = {}
        if alg_varying_params is None:
            alg_varying_params = {}
        if dataset_fixed_params is None:
            dataset_fixed_params = {}
        if dataset_varying_params is None:
            dataset_varying_params = {}

        for alg in self.algorithms:
            alg_name = alg.name

            # Check that every algorithm has an entry in the params dictionary
            if alg_name not in alg_fixed_params:
                alg_fixed_params[alg_name] = {}
            if alg_name not in alg_varying_params:
                alg_varying_params[alg_name] = {}

            # Check that the parameters exist for the algorithm
            for param in alg_fixed_params[alg_name]:
                if param not in alg.parameter_names:
                    raise ValueError(f"Parameter {param} not configured for {alg_name} algorithm.")

            # Convert the parameter iterables to lists
            for param_name in alg_varying_params[alg_name].keys():
                alg_varying_params[alg_name][param_name] = list(alg_varying_params[alg_name][param_name])

        #  Convert parameter iterables to lists
        for param_name in dataset_varying_params.keys():
            dataset_varying_params[param_name] = list(dataset_varying_params[param_name])

        self.alg_fixed_params = alg_fixed_params
        self.alg_varying_params = alg_varying_params
        self.dataset_class = dataset
        self.dataset_fixed_params = dataset_fixed_params
        self.dataset_varying_params = dataset_varying_params
        self.evaluators = evaluators
        self.results_filename = results_filename

        self.results_columns = self.get_results_df_columns()

        # Compute the total number of experiments to run
        num_datasets = 1
        for param_name, values in self.dataset_varying_params.items():
            num_datasets *= len(values)
        self.num_experiments = 0
        for alg_name in self.algorithm_names:
            num_experiments_this_alg = 1
            for param_name, values in self.alg_varying_params[alg_name].items():
                num_experiments_this_alg *= len(values)
            self.num_experiments += num_experiments_this_alg * num_datasets
        self.num_trials = self.num_experiments * self.num_runs

        self.results = None

    def get_results_df_columns(self):
        """Create a list of all the columns in the results file and dataframe."""
        columns = ['trial_id', 'experiment_id', 'run_id', 'algorithm']
        for param_name in self.dataset_fixed_params.keys():
            columns.append(param_name)
        for param_name in self.dataset_varying_params.keys():
            columns.append(param_name)
        for alg_name in self.algorithm_names:
            for param_name in self.alg_fixed_params[alg_name].keys():
                columns.append(param_name)
            for param_name in self.alg_varying_params[alg_name].keys():
                columns.append(param_name)
        columns.append('running_time_s')
        for evaluator in self.evaluators:
            columns.append(evaluator.name)
        return list(OrderedDict.fromkeys(columns))

    def run_all(self, append_results=False) -> alglab.results.Results:
        """Run all the experiments in this suite."""

        # If we are appending the results, make sure that the header of the results file already matches the
        # header we would have written.
        if append_results:
            existing_results = alglab.results.Results(self.results_filename)
            if existing_results.column_names() != self.results_columns:
                raise ValueError("Cannot append results file: column names do not match.")
            true_trial_number = existing_results.results_df.iloc[-1]["trial_id"] + 1
            base_experiment_number = existing_results.results_df.iloc[-1]["experiment_id"] + 1
        else:
            true_trial_number = 1
            base_experiment_number = 1

        reported_trial_number = 1

        file_access_string = 'a' if append_results else 'w'

        with open(self.results_filename, file_access_string) as results_file:
            # Write the header line of the results file
            if not append_results:
                results_file.write(", ".join(self.results_columns))
                results_file.write("\n")

            for run in range(1, self.num_runs + 1):
                experiment_number = base_experiment_number
                for dataset_params in product_dict(**self.dataset_varying_params):
                    resolved_varying_dataset_params = resolve_parameters(self.dataset_fixed_params, dataset_params)
                    full_dataset_params = self.dataset_fixed_params | resolved_varying_dataset_params
                    dataset = self.dataset_class(**full_dataset_params)

                    for alg in self.algorithms:
                        alg_name = alg.name
                        for alg_params in product_dict(**self.alg_varying_params[alg_name]):
                            resolved_varying_alg_params = resolve_parameters(full_dataset_params | self.alg_fixed_params[alg_name], alg_params)
                            full_alg_params = self.alg_fixed_params[alg_name] | resolved_varying_alg_params
                            print(f"Trial {reported_trial_number} / {self.num_trials}: {alg_name} on {dataset} with parameters {full_alg_params}")
                            this_experiment = Experiment(alg, dataset, full_alg_params, self.evaluators)
                            this_experiment.run()

                            this_result = this_experiment.result | full_dataset_params | full_alg_params | \
                                          {'algorithm': alg_name, 'trial_id': true_trial_number,
                                           'experiment_id': experiment_number, 'run_id': run}
                            results_file.write(", ".join([str(this_result[col]) if col in this_result else '' for col in self.results_columns]))
                            results_file.write("\n")
                            results_file.flush()

                            true_trial_number += 1
                            reported_trial_number += 1
                            experiment_number += 1

        # Create a dataframe from the results
        self.results = alglab.results.Results(self.results_filename)
        return self.results



"""Test the evaluation methods of alglab."""
import pytest
import stag.graph
import stag.random
from sklearn.cluster import KMeans
import numpy as np
import alglab.dataset
import alglab.evaluation


# We will use this KMeans implementation throughout the tests.
def kmeans_impl(data: alglab.dataset.PointCloudDataset, k=10):
    sklearn_km = KMeans(n_clusters=k)
    sklearn_km.fit(data.data)
    return sklearn_km.labels_


def test_ari():
    # Create a dataset with ground truth
    data = alglab.dataset.TwoMoonsDataset()

    best_ari = alglab.evaluation.adjusted_rand_index.apply(data, data.gt_labels)
    assert best_ari == 1

    kmeans_labels = kmeans_impl(data, k=2)
    kmeans_ari = alglab.evaluation.adjusted_rand_index.apply(data, kmeans_labels)
    assert 1 > kmeans_ari > 0


def test_ari_no_gt():
    # Create a dataset with no ground truth
    data = alglab.dataset.PointCloudDataset(np.asarray([[1, 2], [2, 3]]))

    # Try to evaluate with some labels
    labels = np.asarray([0, 1])
    with pytest.raises(ValueError, match="ground truth labels"):
        _ = alglab.evaluation.adjusted_rand_index.apply(data, labels)


def test_ari_wrong_dataset_type():
    # Create a non-clusterable dataset
    data = alglab.dataset.NoDataset()

    labels = np.asarray([0, 1])
    with pytest.raises(TypeError, match="dataset type"):
        _ = alglab.evaluation.adjusted_rand_index.apply(data, labels)


def test_ari_graph_dataset():
    # Create a graph dataset
    data = alglab.dataset.SBMDataset(100, 2, 0.5, 0.1)
    gt_labels = data.gt_labels
    ari = alglab.evaluation.adjusted_rand_index.apply(data, gt_labels)
    assert ari == 1


def test_num_vertices():
    num_vertices = alglab.evaluation.num_vertices.apply(alglab.dataset.NoDataset(),
                                                       stag.graph.complete_graph(100))
    assert num_vertices == 100

    # Num vertices works for any alglab dataset
    num_vertices = alglab.evaluation.num_vertices.apply(alglab.dataset.TwoMoonsDataset(),
                                                       stag.graph.complete_graph(100))
    assert num_vertices == 100


def test_avg_degree():
    avg_degree = alglab.evaluation.avg_degree.apply(alglab.dataset.NoDataset(),
                                                   stag.graph.cycle_graph(100))
    assert avg_degree == 2


def test_eigenvalue():
    graph = stag.random.sbm(100, 2, 0.8, 0.01)
    eig = alglab.evaluation.normalised_laplacian_second_eigenvalue.apply(alglab.dataset.NoDataset(),
                                                                        graph)
    assert 0.2 > eig > 0
